import math

def getSegComFromAddress(addr, data=0):
  if addr >= 0xe00:
    addr &= 0xff

  seg = math.floor(addr/2)
  if seg >= 64:
    seg -= 64

  coms = ["0-3", "4-7", "8-11", "12-15"]
  com = addr % 2
  if addr >= 0x80:
    com += 2
  com = coms[com]

  if data > 0:
    rng = com.split("-")
    rng = range(int(rng[0]), int(rng[1])+1)
    com = []
    if data & 0x1:
      com.append(rng[0])
    if data & 0x2:
      com.append(rng[1])
    if data & 0x4:
      com.append(rng[2])
    if data & 0x8:
      com.append(rng[3])
    com = ",".join([str(x) for x in com])

  return { "seg": seg, "com": com }

def printSegCom(addr,data=0):
  segCom = getSegComFromAddress(addr, data)
  return "SEG{},COM{}".format(segCom['seg'], segCom['com'])

# Stolen from MAME... thank you hap from MAME...
seg2x = [
  0,  1,  2,  3,  4,  5,  6,  7,
  35, 8,  9,  10, 11, 12, 13, 14,
  15, 34, 33, 32, 31, 30, 29, 28,
  27, 26, 25, 24, 36, 23, 22, 21,
  20, 19, 18, 17, 16, 37, 38, 39
        ]

def printXY(addr, data=0):
  segCom = getSegComFromAddress(addr, data)
    
  x = seg2x[segCom['seg']]
  y = segCom['com']

  return "X: {}, Y: {}".format(x, y)

# Two return values: the new carry flag and the modified operand
def rrc(x, carry):
  xNew = (carry << 3) | (x >> 1)
  carryNew = (x & 0x1)
  return (xNew, carryNew)

# Technically three return values, but only 2 I care about atm:
# new carry flag and the return value of the addition
# (zero flag is the one I don't care about)
def add(x, y, carry = 0):
  ret = x + y + carry
  carry = ret > 15
  ret = (ret & 0xf)
  return (ret, carry)

def sub(x, y, carry=0):
  ret = (x - y - carry) & 0xf
  carry = x < (y + carry)
  return (ret, carry)

def do_7a5(a, b, carry):
  bSave = b
  b = 0
  (a, carry) = rrc(a, carry)
  (b, carry) = rrc(b, carry)
  m000 = b
  b = bSave
  if b >= 4:
    print("We're zeroizing memory instead. Bail!")
    return
  (a, carry) = add(a, a)
  carry = 0
  (b, carry) = rrc(b, carry)
  (a, carry) = rrc(a, carry)
  # This is an ADC but going into it, carry is always 0. So won't simulate separately.
  (b, carry) = add(b, 6) 
  return [ m000, a, b ]

jumpTargets = { (0,0) : 0x1000, (0,4) : 0x1100, (0,8) : 0x1200, (0,12) : 0x1300,
                (1,0) : 0x1400, (1,4) : 0x1500, (1,8) : 0x1600, (1,12) : 0x1700 }

def do_000(mem):
  m000 = mem[0]
  m001 = mem[1]
  m002 = mem[2]

  a = (m002 & 0x7)
  b = 0
  (a, carry) = add(a, a)
  (a, carry) = add(a, a)
  (b, carry) = add(b, b, carry)
 
  # Also do 0x1020...
  bankPage = jumpTargets[(b,a)]
  return bankPage | (m001 << 4) | m000
  
def getSpriteLoc(a, b, carry):
  return do_000(do_7a5(a, b, carry))

def do_f3d(x, carry=0):
  a = (x & 0xf)
  b = (x & 0xf0) >> 4
  aSave = a
  bSave = b

  (b, carry) = rrc(b, carry)
  (a, carry) = rrc(a, carry)
  (b, carry) = rrc(b, carry)
  (a, carry) = rrc(a, carry)
  b = b & 0x3
  carry = 0
  (lower, carry) = sub(aSave, a, carry)
  (upper, carry) = sub(bSave, b, carry)

  return (upper << 4) | lower

def shouldDiscipline(currentDiscipline, m214, events=None):
  if events:
    events.append("Called with currentDiscipline {}, m214 {}".format(currentDiscipline, m214))
  m214 = currentDiscipline + m214 + 1
  ret = True
  if m214 > 15:
    m214 = (m214 & 0xf)
    ret = False
  return (ret, m214)
